/***************************************************************************
 *cr
 *cr            (C) Copyright 2007 The Board of Trustees of the
 *cr                        University of Illinois
 *cr                         All Rights Reserved
 *cr
 ***************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "cenergy.h"

// Cut off distance
const float g_cutoff = 12.0;

// Initial configuration of binning
//{
#define BIN_LENGTH  4.0
#define BIN_INVLEN  (1.0/BIN_LENGTH)
#define BIN_DEPTH   8
//}

// This function updates global variables as below.
//{

// Dimension of bins
voldim3i binDim;

// Array of bins.
// binBaseAddr: contains surrounding bins
// binZeroAddr: points to starting point of a bin that are in the
//              simulation volume
float4* binBaseAddr;
float4* binZeroAddr;

// Array of number of atoms in bins.
// bincntBaseAddr: contains surrounding bins
// bincntZeroAddr: points to length of starting point of a bin that
//                 are in the simulation volume
int* bincntBaseAddr;
int* bincntZeroAddr;

// Array of extra atoms that are excluded due to capacity of a bin
float4* extraAtoms;
// Number of extra atoms to be handled
int extralen;

voldim3i* NbrList;
int NbrListLen;
//}

void create_uniform_bin(
    voldim3i grid,
    int num_atoms,
    float gridspacing,
    const float* atom,
    float cutoff
  )
{
  int i,j,k;
  int n;

  int num_excluded = 0;

  int lnx = (int) ceilf(grid.x);
  int lny = (int) ceilf(grid.y);
  int lnz = (int) ceilf(grid.z);

  /* create bins */
  /* count extra bins around lattice */
  int c = (int) ceil(cutoff * BIN_INVLEN);
  binDim.x = (int) ceil(lnx * gridspacing * BIN_INVLEN) + 2*c;
  binDim.y = (int) ceil(lny * gridspacing * BIN_INVLEN) + 2*c;
  binDim.z = (int) ceil(lnz * gridspacing * BIN_INVLEN) + 2*c;

  int nbins = binDim.x * binDim.y * binDim.z;
  binBaseAddr = (float4 *) calloc(nbins * BIN_DEPTH, sizeof(float4));
  binZeroAddr = binBaseAddr + ((c * binDim.y + c) * binDim.x + c) * BIN_DEPTH;

  bincntBaseAddr = (int *) calloc(nbins, sizeof(int));
  bincntZeroAddr = bincntBaseAddr + (c * binDim.y + c) * binDim.x + c;

  /* array of extra atoms */
  extraAtoms = (float4 *) calloc(num_atoms, sizeof(float4));

  /*
   * Step 2.1
   *
   * TODO: For given environment, you need to create bins and perform
   *       spatial hashing of atoms onto them. For more information,
   *       please refer to the manual.
   *
   * Output: Fill binZeroAddr and bincntZeroAddr properly.
   *         Fill extraAtoms as needed and set extralen accordingly.
   */

  // Begin of solution

  for (n = 0; n < num_atoms; n++) {
    int n4 = n * 4;

    float4 p;
    p.x = atom[n4 + 0];
    p.y = atom[n4 + 1];
    p.z = atom[n4 + 2];
    p.w = atom[n4 + 3];

    int px = (int) floorf(p.x * BIN_INVLEN);
    int py = (int) floorf(p.y * BIN_INVLEN);
    int pz = (int) floorf(p.z * BIN_INVLEN);

    if (0 <= px && px < binDim.x &&
        0 <= py && py < binDim.y &&
        0 <= pz && pz < binDim.z &&
        p.w != 0) {

      int index = (pz * binDim.y + py) * binDim.x + px;
      float4* bin = &binZeroAddr[index * BIN_DEPTH];
      int numBinItems = bincntZeroAddr[index];
      if (numBinItems < BIN_DEPTH) {
        // copy atom into bin and increase couner for this bin
        bin[numBinItems] = p;
        bincntZeroAddr[index]++;
      } else {
        /* add index to array of extra atoms to be computed with CPU */
        if (extralen >= num_atoms) {
          fprintf(stderr, "exceeded space for storing extra atoms\n");
          return;
        }
        memcpy(&extraAtoms[extralen], &atom[n4], sizeof(float4));
        extralen++;
      }
    } else {
      num_excluded++;
    }
  }

  // End of solution

  printf ("Number of extra atoms: %d\n", extralen);
}

static void calc_extra(
    float* energygrid,
    voldim3i grid,
    float gridspacing,
    float cutoff,
    float k
)
{
  float x,y,z;                     /* Coordinates of current grid point */
  int i,j,m,n;                     /* Loop counters */

  // z is fixed so that it computes over a plane
  z = gridspacing * (float) k;

  for (j = 0; j < grid.y; j++) {
    float y = gridspacing * (float) j;
    long int voxaddr = grid.x*grid.y*k + grid.x*j;

    for (i = 0; i < grid.x; i++) {
      float x = gridspacing * (float) i;

      float e = 0.0;
      for (n=0; n<extralen; n++) {
        float dx = x - extraAtoms[n].x;
        float dy = y - extraAtoms[n].y;
        float dz = z - extraAtoms[n].z;
        float q = extraAtoms[n].w;
  
        float dist = sqrtf(dx*dx + dy*dy + dz*dz);
        if (q == 0.0 || dist > cutoff) continue;

        float t = q / dist;
        float root_s = 1.0 - ((dist * dist) / (cutoff * cutoff));
        e += t * root_s * root_s;
      }

      energygrid[voxaddr + i] += e;
    }
  }
}

// Main interface
//
void calc_energy(
    float *energygrid,
    voldim3i grid,
    const float *atoms,
    int num_atoms,
    float gridspacing,
    float k,
    struct pb_TimerSet* timers)
{
  // Phase 1. Perform the binning process
  create_uniform_bin(grid, num_atoms, gridspacing, atoms, g_cutoff);

  // Phase 2. Create neighbor list
  sol_create_neighbor_list(gridspacing, g_cutoff);

  // Phase 3. Calculate energy using the bins and the neighbor list
  //{
  sol_calc_energy_with_bins(
    energygrid, grid, atoms, num_atoms, gridspacing, g_cutoff, k);

  calc_extra(energygrid, grid, gridspacing, g_cutoff, k);
  //}
}

