/***************************************************************************
 *cr
 *cr            (C) Copyright 2007 The Board of Trustees of the
 *cr                        University of Illinois
 *cr                         All Rights Reserved
 *cr
 ***************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "cenergy.h"

// Cut off distance
const float g_cutoff = 12.0;

// Initial configuration of binning
//{
#define BIN_LENGTH  4.0
#define BIN_INVLEN  (1.0/BIN_LENGTH)
#define BIN_DEPTH   8
//}

// This function updates global variables as below.
//{

// Dimension of bins
voldim3i binDim;

// Array of bins.
// binBaseAddr: contains surrounding bins
// binZeroAddr: points to starting point of a bin that are in the
//              simulation volume
float4* binBaseAddr;
float4* binZeroAddr;

// Array of number of atoms in bins.
// bincntBaseAddr: contains surrounding bins
// bincntZeroAddr: points to length of starting point of a bin that
//                 are in the simulation volume
int* bincntBaseAddr;
int* bincntZeroAddr;

// Array of extra atoms that are excluded due to capacity of a bin
float4* extraAtoms;
// Number of extra atoms to be handled
int extralen;

voldim3i* NbrList;
int NbrListLen;
//}

void create_uniform_bin(
    voldim3i grid,
    int num_atoms,
    float gridspacing,
    const float* atom,
    float cutoff
  )
{
  int lnx = (int) ceilf(grid.x);
  int lny = (int) ceilf(grid.y);
  int lnz = (int) ceilf(grid.z);

  /* create bins */
  /* count extra bins around lattice */
  int c = (int) ceil(cutoff * BIN_INVLEN);
  binDim.x = (int) ceil(lnx * gridspacing * BIN_INVLEN) + 2*c;
  binDim.y = (int) ceil(lny * gridspacing * BIN_INVLEN) + 2*c;
  binDim.z = (int) ceil(lnz * gridspacing * BIN_INVLEN) + 2*c;

  int nbins = binDim.x * binDim.y * binDim.z;
  binBaseAddr = (float4 *) calloc(nbins * BIN_DEPTH, sizeof(float4));
  binZeroAddr = binBaseAddr + ((c * binDim.y + c) * binDim.x + c) * BIN_DEPTH;

  bincntBaseAddr = (int *) calloc(nbins, sizeof(int));
  bincntZeroAddr = bincntBaseAddr + (c * binDim.y + c) * binDim.x + c;

  /* array of extra atoms */
  extraAtoms = (float4 *) calloc(num_atoms, sizeof(float4));

  /*
   * Step 2.1
   *
   * TODO: For given environment, you need to create bins and perform
   *       spatial hashing of atoms onto them. Note that you are free
   *       to modify any part of this function or program. For more
   *       information, please refer to the manual.
   *
   * Output: Fill binZeroAddr and bincntZeroAddr properly.
   *         Fill extraAtoms as needed and set extralen accordingly.
   */

  // Your code here //

  // Begin of solution

  int n;
  for (n = 0; n < num_atoms; n++) {
    int n4 = n * 4;

    float4 p;
    p.x = atom[n4 + 0];
    p.y = atom[n4 + 1];
    p.z = atom[n4 + 2];
    p.w = atom[n4 + 3];

    int px = (int) floorf(p.x * BIN_INVLEN);
    int py = (int) floorf(p.y * BIN_INVLEN);
    int pz = (int) floorf(p.z * BIN_INVLEN);

    if (0 <= px && px < binDim.x &&
        0 <= py && py < binDim.y &&
        0 <= pz && pz < binDim.z &&
        p.w != 0) {

      int index = (pz * binDim.y + py) * binDim.x + px;
      float4* bin = &binZeroAddr[index * BIN_DEPTH];
      int numBinItems = bincntZeroAddr[index];
      if (numBinItems < BIN_DEPTH) {
        // copy atom into bin and increase couner for this bin
        bin[numBinItems] = p;
        bincntZeroAddr[index]++;
      } else {
        /* add index to array of extra atoms to be computed with CPU */
        if (extralen >= num_atoms) {
          fprintf(stderr, "exceeded space for storing extra atoms\n");
          return;
        }
        memcpy(&extraAtoms[extralen], &atom[n4], sizeof(float4));
        extralen++;
      }
    }
  }

  // End of solution

  printf ("Number of extra atoms: %d\n", extralen);
}

/*
 * neighbor list:
 * stored as table of offsets
 * flat index addressing is computed by kernel
 */
void create_neighbor_list(float gridspacing, float cutoff) {
  /* count extra bins around lattice */
  int c = (int) ceil(cutoff * BIN_INVLEN);
  int n = 2 * c + 1;

  NbrList = (voldim3i*) calloc(sizeof(voldim3i), n*n*n);

  /*
   * Step 2.2
   *
   * TODO: Create neighborhood list of bins that represents relative offset
   *       to a regional center. NbrList is the list and NbrListLen is
   *       length of the NbrList. Both are global variables that are to be
   *       updated here. For more information, please refer to the manual.
   *
   * Output: Fill NbrList and set NbrListLen accordingly.
   */

  // Your code here //

  // Begin of solution

  int i, j, k;
  int cnt = 0;
  for (k = -c;  k <= c;  k++) {
    for (j = -c;  j <= c;  j++) {
      for (i = -c;  i <= c;  i++) {
        NbrList[cnt].x = i;
        NbrList[cnt].y = j;
        NbrList[cnt].z = k;
        cnt++;
      }
    }
  }
  NbrListLen = cnt;

  // End of solution

  printf ("Length of neighboring bins list: %d\n", NbrListLen);
}

#define numThreads 8*8*4

__global__ void kernel_calc_energy(float* energygrid, int grid_x, int grid_y,
  int natoms, float gridspacing, const float* atoms, float cutoff, float k,
  int NbrListLen, voldim3i* NbrList, float4* binZeroAddr, int* bincntZeroAddr,
  int dim_x, int dim_y, int dim_z) {

  // Your code here 2/2 //

  // Begin of solution

  float z = gridspacing * (float) k;
  float y = gridspacing * (float) blockIdx.y;
  float x = gridspacing * (float) blockIdx.x;

  int bx = (int) floorf(x * BIN_INVLEN);
  int by = (int) floorf(y * BIN_INVLEN);
  int bz = (int) floorf(z * BIN_INVLEN);

  int i, m;

  __shared__ float e[numThreads];
  __shared__ float t;

  /* thread id */
  int tid = (threadIdx.z*blockDim.y + threadIdx.y)*blockDim.x + threadIdx.x; 
  e[tid] = 0.0;

  /* spatial coordinate of this lattice point */
  for (i = tid; i < NbrListLen; i+=numThreads) {
    int px = bx + NbrList[i].x; 
    int py = by + NbrList[i].y; 
    int pz = bz + NbrList[i].z; 

    if ((0 <= px && px <= dim_x) && 
        (0 <= py && py <= dim_y) &&
        (0 <= pz && pz <= dim_z)) {
      int index = (pz * dim_y + py) * dim_x + px;
      float4* bin = &binZeroAddr[index * BIN_DEPTH];
      int numAtomsInBin = bincntZeroAddr[index];

      for (m = 0; m < numAtomsInBin; m++) {
        float dx = bin[m].x - x;
        float dy = bin[m].y - y;
        float dz = bin[m].z - z;
        float q = bin[m].w;

        float dist = sqrtf(dx*dx + dy*dy + dz*dz);
        if (q == 0.0 || dist > cutoff) continue;

        float t = q / dist;
        float root_s = 1.0 - ((dist * dist) / (cutoff * cutoff));
        e[tid] += t * root_s * root_s;
      }
    }
  }
  __syncthreads();

  if (tid == 0) {
    t = 0.0;
    for (i = 0; i < numThreads; i++) t += e[i];
    long int addr = (unsigned int)((grid_x*grid_y*k) + grid_x*(blockIdx.y))
                     + blockIdx.x;
    energygrid[addr] = t;
  }

  // End of solution
}

void calc_energy_with_bins_on_gpu(
    float *energygrid,
    voldim3i grid_dim,
    const float *atoms,
    int num_atoms,
    float gridspacing,
    float cutoff,
    float k,
    struct pb_TimerSet* timers)
{
  pb_SwitchToTimer(timers, pb_TimerID_COPY);

  /* count extra bins around lattice */
  int c = (int) ceil(cutoff * BIN_INVLEN);
  int nbins = binDim.x * binDim.y * binDim.z;

  float* dev_energygrid;
  float* dev_atoms;

  voldim3i *dev_NbrList;
 
  float4* dev_binBaseAddr;
  float4* dev_binZeroAddr;

  int* dev_bincntBaseAddr;
  int* dev_bincntZeroAddr;

  dim3 dg;
  dim3 db;
  /*
   * Step 3.1
   *
   * TODO: 
   *
   * Output: Fill energygrid.
   */

  // Your code here 1/2 //

  // Begin of solution

  cudaMalloc((void**)&dev_energygrid, sizeof(float) * grid_dim.x * grid_dim.y);
  cudaMemset(dev_energygrid, 0, sizeof(float) * grid_dim.x * grid_dim.y);
  cudaMalloc((void**)&dev_atoms, sizeof(float) * 4 * num_atoms);
  cudaMemcpy(dev_atoms, atoms, sizeof(float) * 4 * num_atoms,
              cudaMemcpyHostToDevice);

  cudaMalloc((void**)&dev_NbrList, sizeof(voldim3i) * NbrListLen);
  cudaMemcpy(dev_NbrList, NbrList, sizeof(voldim3i) * NbrListLen,
              cudaMemcpyHostToDevice);

  cudaMalloc((void**)&dev_binBaseAddr, nbins * BIN_DEPTH * sizeof(float4));
  cudaMemcpy(dev_binBaseAddr, binBaseAddr,
              nbins * BIN_DEPTH * sizeof(float4), cudaMemcpyHostToDevice);
  dev_binZeroAddr = dev_binBaseAddr
                    + ((c * binDim.y + c) * binDim.x + c) * BIN_DEPTH;

  cudaMalloc((void**)&dev_bincntBaseAddr, nbins * sizeof(int));
  cudaMemcpy(dev_bincntBaseAddr, bincntBaseAddr, nbins * sizeof(int),
              cudaMemcpyHostToDevice);
  dev_bincntZeroAddr = dev_bincntBaseAddr
                       + (c * binDim.y + c) * binDim.x + c;

  // Begin of solution

  // Kernel call here
  //{
  // CUDA grid & block configuration
  dg.x = grid_dim.x; dg.y = grid_dim.y; dg.z = 1;
  db.x = 8; db.y = 8; db.z = 4;

  pb_SwitchToTimer(timers, pb_TimerID_GPU);

  kernel_calc_energy<<<dg, db>>>(
    dev_energygrid, grid_dim.x, grid_dim.y, num_atoms,
    gridspacing, dev_atoms, cutoff, k,
    NbrListLen, dev_NbrList, dev_binZeroAddr, dev_bincntZeroAddr,
    binDim.x, binDim.y, binDim.z
  );
  cudaThreadSynchronize();
  if(cudaGetLastError()) {
    printf ("Kernel call failed.\n");
  }
  //}

  pb_SwitchToTimer(timers, pb_TimerID_COPY);

  // Get result back to the system memory
  cudaMemcpy(
      energygrid, dev_energygrid, sizeof(float) * grid_dim.x * grid_dim.y,
      cudaMemcpyDeviceToHost);
}

static void calc_extra(
  float* energygrid,
  voldim3i grid_dim,
  float gridspacing,
  float cutoff,
  float k,
  pb_TimerSet* timers)
{
  pb_SwitchToTimer(timers, pb_TimerID_COMPUTE);

  int i,j,n;
  float x,y,z;
  z = gridspacing * (float) k;

  for (j = 0; j < grid_dim.y; j++) {
    y = gridspacing * (float) j;
    long int voxaddr = grid_dim.x*grid_dim.y*k + grid_dim.x*j;

    for (i = 0; i < grid_dim.x; i++) {
      x = gridspacing * (float) i;

      float e = 0.0;
      for (n=0; n<extralen; n++) {
        float dx = x - extraAtoms[n].x;
        float dy = y - extraAtoms[n].y;
        float dz = z - extraAtoms[n].z;
        float q = extraAtoms[n].w;
        float dist = sqrtf(dx*dx + dy*dy + dz*dz);

        if (q == 0.0 || dist > cutoff) continue;
        float t = q / dist;
        float root_s = 1.0 - ((dist * dist) / (cutoff * cutoff));
        e += t * root_s * root_s;
      }

      energygrid[voxaddr + i] += e;
    }
  }
}

void calc_energy(
    float *energygrid,
    voldim3i grid_dim,
    const float *atoms,
    int num_atoms,
    float gridspacing,
    float k,
    struct pb_TimerSet* timers)
{
  pb_SwitchToTimer(timers, pb_TimerID_COMPUTE);

  // Phase 1. Perform the binning process
  create_uniform_bin(grid_dim, num_atoms, gridspacing, atoms, g_cutoff);

  // Phase 2. Create the neighbor list 
  create_neighbor_list(gridspacing, g_cutoff);

  // Phase 3. Calculate energy using the bins and the neighbor list
  //{
  calc_energy_with_bins_on_gpu(
    energygrid, grid_dim, atoms, num_atoms, gridspacing, g_cutoff, k, timers);

  calc_extra(energygrid, grid_dim, gridspacing, g_cutoff, k, timers);
  //}
}

