/***************************************************************************
 *cr
 *cr            (C) Copyright 2007 The Board of Trustees of the
 *cr                        University of Illinois
 *cr                         All Rights Reserved
 *cr
 ***************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "cenergy.h"

// Cut off distance
extern float cutoff;

// Initial configuration of binning
//{
#define BIN_LENGTH  4.0
#define BIN_INVLEN  (1.0/BIN_LENGTH)
#define BIN_DEPTH   8
//}

// This function updates global variables as below.
//{

// Dimension of bins
extern voldim3i binDim;

// Array of bins.
// binBaseAddr: contains surrounding bins
// binZeroAddr: points to starting point of a bin that are in the
//              simulation volume
extern float4* binBaseAddr;
extern float4* binZeroAddr;

// Array of number of atoms in bins.
// bincntBaseAddr: contains surrounding bins
// bincntZeroAddr: points to length of starting point of a bin that
//                 are in the simulation volume
extern int* bincntBaseAddr;
extern int* bincntZeroAddr;

// Array of extra atoms that are excluded due to capacity of a bin
extern float4* extraAtoms;
// Number of extra atoms to be handled
extern int extralen;

extern voldim3i* NbrList;
extern int NbrListLen;
//}

#define numThreads (8*4*8)

static __global__ void kernel_calc_energy(float* energygrid, int grid_x, int grid_y,
  int natoms, float gridspacing, const float* atoms, float cutoff, float k,
  int NbrListLen, voldim3i* NbrList, float4* binZeroAddr, int* bincntZeroAddr,
  int dim_x, int dim_y, int dim_z)
{

  float z = gridspacing * (float) k;
  float y = gridspacing * (float) blockIdx.y;
  float x = gridspacing * (float) blockIdx.x;

  int bx = (int) floorf(x * BIN_INVLEN);
  int by = (int) floorf(y * BIN_INVLEN);
  int bz = (int) floorf(z * BIN_INVLEN);

  int i, m;

  __shared__ float e[numThreads];
  __shared__ float t;

  /* thread id */
  int tid = (threadIdx.z*blockDim.y + threadIdx.y)*blockDim.x + threadIdx.x; 
  e[tid] = 0.0f;

  /* spatial coordinate of this lattice point */
  for (i = tid; i < NbrListLen; i+=numThreads) {
    int px = bx + NbrList[i].x; 
    int py = by + NbrList[i].y; 
    int pz = bz + NbrList[i].z; 

    if ((0 <= px && px <= dim_x) && 
        (0 <= py && py <= dim_y) &&
        (0 <= pz && pz <= dim_z)) {
      int index = (pz * dim_y + py) * dim_x + px;
      float4* bin = &binZeroAddr[index * BIN_DEPTH];
      int numAtomsInBin = bincntZeroAddr[index];

      for (m = 0; m < numAtomsInBin; m++) {
        float dx = bin[m].x - x;
        float dy = bin[m].y - y;
        float dz = bin[m].z - z;
        float q = bin[m].w;

        float dist = sqrtf(dx*dx + dy*dy + dz*dz);
        if (q == 0.0f || dist > cutoff) continue;

        float t = q / dist;
        float root_s = 1.0f - ((dist * dist) / (cutoff * cutoff));
        e[tid] += t * root_s * root_s;
      }
    }
  }
  __syncthreads();

  if (tid == 0) {
    t = 0.0f;
    for (i = 0; i < numThreads; i++) t += e[i];
    long int addr = (unsigned int)((grid_x*grid_y*k) + grid_x*(blockIdx.y))
                     + blockIdx.x;
    energygrid[addr] = t;
  }
  __syncthreads();
}

void sol_calc_energy_with_bins_on_gpu(
    float *energygrid,
    voldim3i grid,
    const float *atoms,
    int num_atoms,
    float gridspacing,
    float cutoff,
    float k,
    struct pb_TimerSet* timers)
{
  // CUDA programming begins here
  pb_SwitchToTimer(timers, pb_TimerID_COPY);

  /* count extra bins around lattice */
  int c = (int) ceil(cutoff * BIN_INVLEN);
  int nbins = binDim.x * binDim.y * binDim.z;

  float* dev_energygrid;
  float* dev_atoms;
  cudaMalloc((void**)&dev_energygrid, sizeof(float) * grid.x * grid.y);
  cudaMemset(dev_energygrid, 0, sizeof(float) * grid.x * grid.y);
  cudaMalloc((void**)&dev_atoms, sizeof(float) * 4 * num_atoms);
  cudaMemcpy(dev_atoms, atoms, sizeof(float) * 4 * num_atoms,
              cudaMemcpyHostToDevice);

  voldim3i *dev_NbrList;
  cudaMalloc((void**)&dev_NbrList, sizeof(voldim3i) * NbrListLen);
  cudaMemcpy(dev_NbrList, NbrList, sizeof(voldim3i) * NbrListLen,
              cudaMemcpyHostToDevice);

  float4* dev_binBaseAddr;
  float4* dev_binZeroAddr;
  cudaMalloc((void**)&dev_binBaseAddr, nbins * BIN_DEPTH * sizeof(float4));
  cudaMemcpy(dev_binBaseAddr, binBaseAddr,
              nbins * BIN_DEPTH * sizeof(float4), cudaMemcpyHostToDevice);
  dev_binZeroAddr = dev_binBaseAddr
                    + ((c * binDim.y + c) * binDim.x + c) * BIN_DEPTH;

  int* dev_bincntBaseAddr;
  int* dev_bincntZeroAddr;
  cudaMalloc((void**)&dev_bincntBaseAddr, nbins * sizeof(int));
  cudaMemcpy(dev_bincntBaseAddr, bincntBaseAddr, nbins * sizeof(int),
              cudaMemcpyHostToDevice);
  dev_bincntZeroAddr = dev_bincntBaseAddr
                       + (c * binDim.y + c) * binDim.x + c;

  // CUDA grid & block configuration
  dim3 dg;
  dim3 db;
  dg.x = grid.x; dg.y = grid.y; dg.z = 1;
  db.x = 8; db.y = 4; db.z = 8;

  // Kernel call here
  //{
  pb_SwitchToTimer(timers, pb_TimerID_GPU);

  kernel_calc_energy<<<dg, db>>>(
    dev_energygrid, grid.x, grid.y, num_atoms,
    gridspacing, dev_atoms, cutoff, k,
    NbrListLen, dev_NbrList, dev_binZeroAddr, dev_bincntZeroAddr,
    binDim.x, binDim.y, binDim.z
  );
  if(cudaGetLastError()) {
    printf ("Kernel call failed.\n");
  }
  //}

  pb_SwitchToTimer(timers, pb_TimerID_COPY);

  // Get result back to the system memory
  cudaMemcpy(energygrid, dev_energygrid, sizeof(float) * grid.x * grid.y,
              cudaMemcpyDeviceToHost);

}

