#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <cuda.h>
#include <parboil.h>

#include "UDTypes.h"
#include "WKFUtils.h"
#include "CUDA_interface.h"
#include "CPU_kernels.h"

#define PI 3.14159265

/************************************************************ 
 * This function reads the parameters from the file provided
 * as a comman line argument.
 ************************************************************/
void setParameters(FILE* file, parameters* p){
  fscanf(file,"aquisition.numsamples=%d\n",&(p->numSamples));
  fscanf(file,"aquisition.kmax=%f\n",&(p->kMax));
  fscanf(file,"aquisition.matrixSize=%d %d %d\n", &(p->aquisitionMatrixSize[0]), &(p->aquisitionMatrixSize[1]), &(p->aquisitionMatrixSize[2]));
  fscanf(file,"reconstruction.matrixSize=%d %d %d\n", &(p->reconstructionMatrixSize[0]), &(p->reconstructionMatrixSize[1]), &(p->reconstructionMatrixSize[2]));
  fscanf(file,"gridding.matrixSize=%d %d %d\n", &(p->gridSize[0]), &(p->gridSize[1]), &(p->gridSize[2]));
  fscanf(file,"gridding.oversampling=%f\n", &(p->oversample));
  fscanf(file,"kernel.width=%f\n", &(p->kernelWidth));
  fscanf(file,"kernel.useLUT=%d\n", &(p->useLUT));

  printf("  Number of samples = %d\n", p->numSamples);
  printf("  Grid Size = %dx%dx%d\n", p->gridSize[0], p->gridSize[1], p->gridSize[2]);
  printf("  Kernel Width = %f\n", p->kernelWidth);
  printf("  GPU Binsize = %d\n", p->binsize);
}

/************************************************************ 
 * This function reads the sample point data from the kspace
 * and klocation files (and sdc file if provided) into the
 * sample array.
 * Returns the number of samples it successfully read.
 ************************************************************/
unsigned int readSampleData(int n, FILE* uksdata_f, ReconstructionSample* samples){
  unsigned int i;

  for(i=0; i<n; i++){
    if (feof(uksdata_f)){
      break;
    }
    fread((void*) &(samples[i]), sizeof(float),2, uksdata_f);
  }

  return i;
}

float rand_gauss(float dev)
{
	static float y1;
	static float y2;
	static unsigned int i=0;
	float x1,x2,w;

	if((i&0x1)==1)
	{
		i++;
		return y2*dev;
	}
	else
	{
		do
		{
			x1=((float)rand())/((float)RAND_MAX);
			x2=((float)rand())/((float)RAND_MAX);

			x1 = 2.0*x1-1.0;
			x2 = 2.0*x2-1.0;

			w = (x1*x1)+(x2*x2);
		
		}while(w>=1.0 || w==0.0);

		w=sqrt((-2.0*log(w))/w);
		y1=x1*w;
		y2=x2*w;

		i++;

		return y1*dev;
	}
}

int main (int argc, char* argv[]){
  
  struct pb_Parameters* prms;
  prms = pb_ReadParameters(&argc,argv);

  char uksdata[250];
  parameters params;

  float dev;

  if(argc==3)
  {
	params.binsize=atoi(argv[1]);
	dev=atof(argv[2]);
	params.sync=0;
  }
  else
  {
	printf("Error: insufficient arguments!\n");
	exit(1);
  }

  FILE* uksfile_f = NULL;
  FILE* uksdata_f = NULL;

  strcpy(uksdata,prms->inpFiles[0]);
  strcat(uksdata,".data");

  uksfile_f = fopen(prms->inpFiles[0],"r");
  if (uksfile_f == NULL){
    printf("ERROR: Could not open %s\n",prms->inpFiles[0]);
    exit(1);
  }

  printf("\nReading parameters\n");

  setParameters(uksfile_f, &params);

  printf("  Stddev: %.2f\n",dev);

  ReconstructionSample* samples; //Input Data
  float* LUT; //use look-up table for faster execution on CPU (intermediate data)
  unsigned int sizeLUT; //set in the function calculateLUT (intermediate data)

  cmplx* gridData; //Output Data
  float* sampleDensity; //Output Data

  cudaMallocHost((void**)&samples, params.numSamples*sizeof(ReconstructionSample));
  if (samples == NULL){
    printf("ERROR: Unable to allocate memory for input data\n");
    exit(1);
  }

  uksdata_f = fopen(uksdata,"rb");

  if(uksdata_f == NULL){
    printf("ERROR: Could not open data file\n");
    exit(1);
  }

  printf("Reading input data from files\n");

  unsigned int n = readSampleData(params.numSamples, uksdata_f, samples);
  fclose(uksdata_f);

  float kScale0 = float(params.aquisitionMatrixSize[0])/(float(params.reconstructionMatrixSize[0])*float(params.kMax));
  float a = kScale0*(params.gridSize[0]-1.0)/2.0;
  float b = (params.gridSize[0]-1.0)/2.0;

  srand(54321);
  for(int i=0;i<n;i++)
  {
	do
	{
		samples[i].kX=rand_gauss(dev/a)+params.gridSize[0]/(2*a)-b/a;
	}while(samples[i].kX<(-b/a) || samples[i].kX>=((params.gridSize[0]-b)/a));

 	do
        {
                samples[i].kY=rand_gauss(dev/a)+params.gridSize[1]/(2*a)-b/a;
        }while(samples[i].kY<(-b/a) || samples[i].kY>=((params.gridSize[1]-b)/a));

 	do
        {
                samples[i].kZ=rand_gauss(dev/a)+params.gridSize[2]/(2*a)-b/a;
        }while(samples[i].kZ<(-b/a) || samples[i].kZ>=((params.gridSize[2]-b)/a));
	samples[i].sdc=1.0;
  }

  if (params.useLUT){
    printf("Generating Look-Up Table\n");
    float beta = PI * sqrt(4*params.kernelWidth*params.kernelWidth/(params.oversample*params.oversample) * (params.oversample-.5)*(params.oversample-.5)-.8);
    calculateLUT(beta, params.kernelWidth, LUT, sizeLUT);
  }

  int gridNumElems = params.gridSize[0] * params.gridSize[1] * params.gridSize[2];

  wkf_timerhandle timer = wkf_timer_create();

  cudaMallocHost((void**)&gridData, gridNumElems*sizeof(cmplx));
  cudaMallocHost((void**)&sampleDensity, gridNumElems*sizeof(float));
  if (sampleDensity == NULL || gridData == NULL){
    printf("ERROR: Unable to allocate memory for output data\n");
    exit(1);
  }

  printf("Running CUDA version\n");
  wkf_timer_start(timer);

  //Interface function to GPU implementation of gridding
  CUDA_interface(n, params, samples, LUT, sizeLUT, gridData, sampleDensity);

  wkf_timer_stop(timer);
  printf("  Total CUDA runtime is %f\n", wkf_timer_time(timer));
  wkf_timer_destroy(timer);

  FILE* outfile;
  if(!(outfile=fopen(prms->outFile,"w")))
  {
	printf("Cannot open output file!\n");
  }

  int t = 2;
  fwrite(&t,sizeof(int),1,outfile);
  t=gridNumElems;
  fwrite(&t,sizeof(int),1,outfile);

  for(int i=0;i<gridNumElems;i++)
  {
	fwrite(&gridData[i],sizeof(cmplx),1,outfile);
	fwrite(&sampleDensity[i],sizeof(float),1,outfile);
  }

  fclose(outfile);

  if (params.useLUT){
    free(LUT);
  }
  cudaFreeHost(samples);
  cudaFreeHost(gridData);
  cudaFreeHost(sampleDensity);

  return 0;
}
