#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#define NUM_BANKS 16
#define LOG_NUM_BANKS 4

#define CONFLICT_FREE_OFFSET(index) ((index) >> LOG_NUM_BANKS + (index) >> (2*LOG_NUM_BANKS))
#define MAXIMUM(x,y) (x>y)?x:y

////////////////////////////////////////////////////////////////////////////////
// Kernels
////////////////////////////////////////////////////////////////////////////////
__global__ void scan_L1_kernel(int n, int* data, int* inter)
{
    __shared__ int s_data[1088]; // 1088 = 1024 + (1024/16);

    int thid = threadIdx.x;
    int g_ai = blockIdx.x*2*blockDim.x + threadIdx.x;
    int g_bi = g_ai + blockDim.x;

    int s_ai = thid;
    int s_bi = thid + blockDim.x;

    s_ai += CONFLICT_FREE_OFFSET(s_ai);
    s_bi += CONFLICT_FREE_OFFSET(s_bi);

    s_data[s_ai] = (g_ai < n) ? data[g_ai] : 0;
    s_data[s_bi] = (g_bi < n) ? data[g_bi] : 0;

    int stride = 1;
    for (int d = blockDim.x; d > 0; d >>= 1)
    {
        __syncthreads();

        if (thid < d)
        {
            int i  = 2*stride*thid;
            int ai = i + stride - 1;
            int bi = ai + stride;

            ai += CONFLICT_FREE_OFFSET(ai);
            bi += CONFLICT_FREE_OFFSET(bi);

            s_data[bi] = MAXIMUM(s_data[ai], s_data[bi]);
        }

        stride *= 2;
    }

    if (thid == 0){
        int last = blockDim.x*2 -1;
        last += CONFLICT_FREE_OFFSET(last);
        inter[blockIdx.x] = s_data[last];
        s_data[last] = 0;
    }

    for (int d = 1; d <= blockDim.x; d *= 2)
    {
        stride >>= 1;

        __syncthreads();

        if (thid < d)
        {
            int i  = 2*stride*thid;
            int ai = i + stride - 1;
            int bi = ai + stride;

            ai += CONFLICT_FREE_OFFSET(ai);
            bi += CONFLICT_FREE_OFFSET(bi);

            int t  = s_data[ai];
            s_data[ai] = s_data[bi];
            s_data[bi] = MAXIMUM(t,s_data[bi]);
        }
    }
    __syncthreads();

    if (g_ai < n) { data[g_ai] = s_data[s_ai]; }
    if (g_bi < n) { data[g_bi] = s_data[s_bi]; }
}

__global__ void scan_inter1_kernel(int* data, int iter)
{
    __shared__ int s_data[544]; // 544 = 512 + (512/16);

    int thid = threadIdx.x;
    int gthid = (blockIdx.x*blockDim.x + threadIdx.x);
    int gi = 2*iter*gthid;
    int g_ai = gi + iter - 1;
    int g_bi = g_ai + iter;

    int s_ai = 2*thid;
    int s_bi = 2*thid + 1;

    s_ai += CONFLICT_FREE_OFFSET(s_ai);
    s_bi += CONFLICT_FREE_OFFSET(s_bi);

    s_data[s_ai] = data[g_ai];
    s_data[s_bi] = data[g_bi];

    int stride = 1;
    for (int d = blockDim.x; d > 0; d >>= 1)
    {
        __syncthreads();

        if (thid < d)
        {
            int i  = 2*stride*thid;
            int ai = i + stride - 1;
            int bi = ai + stride;

            ai += CONFLICT_FREE_OFFSET(ai);
            bi += CONFLICT_FREE_OFFSET(bi);

            s_data[bi] = MAXIMUM(s_data[ai], s_data[bi]);
        }

        stride *= 2;
    }

    data[g_ai] = s_data[s_ai];
    data[g_bi] = s_data[s_bi];
}

__global__ void scan_inter2_kernel(int* data, int iter)
{
    __shared__ int s_data[544]; // 544 = 512 + (512/16);

    int thid = threadIdx.x;
    int gthid = (blockIdx.x*blockDim.x + threadIdx.x);
    int gi = 2*iter*gthid;
    int g_ai = gi + iter - 1;
    int g_bi = g_ai + iter;

    int s_ai = 2*thid;
    int s_bi = 2*thid + 1;

    s_ai += CONFLICT_FREE_OFFSET(s_ai);
    s_bi += CONFLICT_FREE_OFFSET(s_bi);

    s_data[s_ai] = data[g_ai];
    s_data[s_bi] = data[g_bi];

    int stride = 512;
    for (int d = 1; d <= blockDim.x; d *= 2)
    {
        stride >>= 1;

        __syncthreads();

        if (thid < d)
        {
            int i  = 2*stride*thid;
            int ai = i + stride - 1;
            int bi = ai + stride;

            ai += CONFLICT_FREE_OFFSET(ai);
            bi += CONFLICT_FREE_OFFSET(bi);

            int t  = s_data[ai];
            s_data[ai] = s_data[bi];
            s_data[bi] = MAXIMUM(t,s_data[bi]);
        }
    }
    __syncthreads();

    data[g_ai] = s_data[s_ai];
    data[g_bi] = s_data[s_bi];
}

__global__ void uniformAdd(int n, int *data, int *inter)
{

    __shared__ int uni;
    if (threadIdx.x == 0) { uni = inter[blockIdx.x]; }
    __syncthreads();

    int g_ai = blockIdx.x*2*blockDim.x + threadIdx.x;
    int g_bi = g_ai + blockDim.x;

    if (g_ai < n) { data[g_ai] = MAXIMUM(uni,data[g_ai]); }
    if (g_bi < n) { data[g_bi] = MAXIMUM(uni,data[g_bi]); }
}

void scanLargeArray( int gridNumElements, int* data_d) {
    unsigned int gridNumElems = gridNumElements;    

    // allocate device memory input and output arrays
    int* inter_d = NULL;

    cudaMalloc( (void**) &inter_d, 512*512*sizeof(int));
    cudaMemset (inter_d, 0, 512*512*sizeof(int));
    if(cudaGetLastError());    

    // Run the prescan
    int size = (gridNumElems+1023)/1024;

    for (int i=0; i < (size+65534)/65535; i++){
        int gridSize = ((size-(i*65535)) > 65535) ? 65535 : (size-i*65535);
        int numElems = ((gridNumElems-(i*65535*1024)) > (65535*1024)) ? (65535*1024) : (gridNumElems-(i*65535*1024));

        dim3 block2 (512);
        dim3 grid2 (gridSize);
        scan_L1_kernel<<<grid2, block2>>>(numElems, data_d+(i*65535*1024), inter_d+(i*65535));
	if(cudaGetLastError())
        	printf("Unable to launch scan_L1_kernel\n");
    }

    int inter_size = 512*512;
    int stride = 1;
    for (int d = inter_size; d > 1; d /= 512)
    {
        dim3 block3 (256);
        dim3 grid3 (d/512);
        scan_inter1_kernel<<<grid3, block3>>>(inter_d, stride);
        if(cudaGetLastError())
		printf("Unable to launch scan_inter1_kernel\n");
        stride *= 512;
    }

    cudaMemset(&(inter_d[inter_size-1]), 0, sizeof(int));

    for (int d = 512; d <= inter_size; d *= 512)
    {
        stride /= 512;
        dim3 block4 (256);
        dim3 grid4 (d/512);
        scan_inter2_kernel<<<grid4, block4>>>(inter_d, stride);
	if(cudaGetLastError())
        	printf("Unable to launch scan_inter2_kernel\n");
    }

    for (int i=0; i < (size+65534)/65535; i++){
        int gridSize = ((size-(i*65535)) > 65535) ? 65535 : (size-i*65535);
        int numElems = ((gridNumElems-(i*65535*1024)) > (65535*1024)) ? (65535*1024) : (gridNumElems-(i*65535*1024));

        dim3 block5 (512);
        dim3 grid5 (gridSize);
        uniformAdd<<<grid5, block5>>>(numElems, data_d+(i*65535*1024), inter_d+(i*65535));
	if(cudaGetLastError())
        	printf("Unable to launch uniformAdd kernel\n");
    }

    cudaThreadSynchronize();

    // cleanup memory
    cudaFree( inter_d);
}
