
#include "common.h"

#define BLOCK_SIZE_X 16
#define BLOCK_SIZE_Y 16
#define COARSEN_FACTOR 2



__global__ void block2D_naive(float fac,float *A0,float *Anext, int nx, int ny, int nz)
{
	
	int i = blockIdx.x*blockDim.x+threadIdx.x;
	int j = blockIdx.y*blockDim.y+threadIdx.y;
	if( i>0 && j>0 &&(i<nx-1) &&(j<ny-1) )
	{
		for(int k=1;k<nz-1;k++)
		{
			Anext[Index3D (nx, ny, i, j, k)] = 
			A0[Index3D (nx, ny, i, j, k + 1)] +
			A0[Index3D (nx, ny, i, j, k - 1)] +
			A0[Index3D (nx, ny, i, j + 1, k)] +
			A0[Index3D (nx, ny, i, j - 1, k)] +
			A0[Index3D (nx, ny, i + 1, j, k)] +
			A0[Index3D (nx, ny, i - 1, j, k)]
			- 6.0f * A0[Index3D (nx, ny, i, j, k)] / (fac*fac);
		}
	}

}


//For Optimization 1
__global__ void block2D_shared(float fac,float *A0,float *Anext, int nx, int ny, int nz)
{
	
	int inx=threadIdx.x;
	int iny=threadIdx.y;
	int i = blockIdx.x*blockDim.x+inx;
    int j = blockIdx.y*blockDim.y+iny;
	
	//linearize thread id
	int id=iny*BLOCK_SIZE_X+inx;
	
	//reshape to shared memory thread ids
	int snx=id%HALO_SIZE_X;
	int sny=id/HALO_SIZE_X;
	int si=blockIdx.x*blockDim.x+snx-1;
	int sj=blockIdx.y*blockDim.y+sny-1;
	
	//malloc shared memory
	__shared__ float sh_A0[HALO_SIZE_Y][HALO_SIZE_X];
	
	sh_A0[iny][inx]=0.0f;
	__syncthreads();
	
	//get active thread ids
	bool thread_active = id<HALF_TOTAL_HALO_SIZE;
	
	//calculate reasonable region for load and store
	bool w_region =  i>0 && j>0 &&(i<nx-1) &&(j<ny-1) ;
	bool r_region =  si>=0 && sj>=0  &&(si<nx) &&(sj<ny) ;
	bool r_region_2 =  si>=0 &&(si<nx) &&((sj+HALF_HALO_SIZE_Y)<ny) ;
	
	

	for(int k=1;k<nz-1;k++)
	{
	
		//load twice to collect whole data
		//because thread number (16*16) is smaller than data size(18*18)
		if(thread_active)
		{
			if(r_region)
			{
				sh_A0[sny][snx]=A0[Index3D (nx, ny, si, sj, k)];
			}
			if(r_region_2)
			{
				sh_A0[sny+HALF_HALO_SIZE_Y][snx]=A0[Index3D (nx, ny, si, sj+HALF_HALO_SIZE_Y, k)];
			}
		}
		__syncthreads();

		if(w_region)
		{
			Anext[Index3D (nx, ny, i, j, k)] =	
			A0[Index3D (nx, ny, i, j, k + 1)] +
			A0[Index3D (nx, ny, i, j, k - 1)] +
			sh_A0[iny+2][inx+1]  +
			sh_A0[iny][inx+1]  +
			sh_A0[iny+1][inx+2]  +
			sh_A0[iny+1][inx] 
			- 6.0f * sh_A0[iny+1][inx+1] / (fac*fac);
		}
		__syncthreads();
	}
}


//For Optimization 1, alternative 
__global__ void block2D_hybrid(float fac,float *A0,float *Anext, int nx, int ny, int nz)
{
	int inx=threadIdx.x;
	int iny=threadIdx.y;
	int i = blockIdx.x*blockDim.x+inx;
    int j = blockIdx.y*blockDim.y+iny;
	
	//malloc shared memory
	__shared__ float sh_A0[BLOCK_SIZE_Y][BLOCK_SIZE_X];
	sh_A0[iny][inx]=0.0f;
	__syncthreads();

	
	//calculate reasonable region
	bool w_region =  i>0 && j>0 &&(i<nx-1) &&(j<ny-1) ;
	bool x_l_bound = (inx==0);
	bool x_h_bound = (inx==(blockDim.x-1));
	bool y_l_bound = (iny==0);
	bool y_h_bound = (iny==(blockDim.y-1));
	
	
	for(int k=1;k<nz-1;k++)
	{

		sh_A0[iny][inx]=A0[Index3D (nx, ny, i, j, k)];
		__syncthreads();
	
		
		if(w_region)
		{
			//calculate partial results of z direction
			float partial=A0[Index3D (nx, ny, i, j, k + 1)] 
						 +A0[Index3D (nx, ny, i, j, k - 1)] 
						 - 6.0f * sh_A0[iny][inx]  / (fac*fac);
			
			//load data from xy plane
			//if data in shared memory load from shared memory
			//if not, load data from global memory
			float a_left,a_right,a_top,a_down;
			if(x_l_bound)
			{
				a_left=A0[Index3D (nx, ny, i-1, j, k )];
			}
			else
			{
				a_left=sh_A0[iny][inx-1];
			}
			
			if(x_h_bound)
			{
				a_right=A0[Index3D (nx, ny, i+1, j, k )];
			}
			else
			{
				a_right=sh_A0[iny][inx+1];
			}
				
			if(y_l_bound)
			{
				a_down=A0[Index3D (nx, ny, i, j-1, k )];
			}
			else
			{
				a_down=sh_A0[iny-1][inx];
			}
			
			if(y_h_bound)
			{
				a_top=A0[Index3D (nx, ny, i, j+1, k )];
			}
			else
			{
				a_top=sh_A0[iny+1][inx];
			}
	
			//calculate total results
			Anext[Index3D (nx, ny, i, j, k)] =	a_left + a_right + a_top + a_down + partial;

		}
		__syncthreads();
	}
}


//For Optimization 2 
__global__ void block2D_hybrid_swap(float fac,float *A0,float *Anext, int nx, int ny, int nz)
{
	int inx=threadIdx.x;
	int iny=threadIdx.y;
	int i = blockIdx.x*blockDim.x+inx;
    int j = blockIdx.y*blockDim.y+iny;
	//malloc shared memory (Only for current center region)
	__shared__ float sh_A0[BLOCK_SIZE_Y][BLOCK_SIZE_X];
	sh_A0[iny][inx]=0.0f;
	__syncthreads();

	//get reasonable region for load and store
	bool w_region =  i>0 && j>0 &&(i<nx-1) &&(j<ny-1) ;
	bool x_l_bound = (inx==0);
	bool x_h_bound = (inx==(blockDim.x-1));
	bool y_l_bound = (iny==0);
	bool y_h_bound = (iny==(blockDim.y-1));
	
	
	//registers for top and bottom planes
	float bottom,top;
	//load bottom and current planes
	bottom=A0[Index3D (nx, ny, i, j, 0)];
	sh_A0[iny][inx]=A0[Index3D (nx, ny, i, j, 1)];
	__syncthreads();
	
	for(int k=1;k<nz-1;k++)
	{

		//load top plane
		top=A0[Index3D (nx, ny, i, j, k+1)];

		if(w_region)
		{
			//get necessary data on xy plane (current plane)
			//if it in shared memory, get it from shared memory
			//if not, load it from global
			float a_left,a_right,a_top,a_down;
			if(x_l_bound)
			{
				a_left=A0[Index3D (nx, ny, i-1, j, k )];
			}
			else
			{
				a_left=sh_A0[iny][inx-1];
			}
			
			if(x_h_bound)
			{
				a_right=A0[Index3D (nx, ny, i+1, j, k )];
			}
			else
			{
				a_right=sh_A0[iny][inx+1];
			}
				
			if(y_l_bound)
			{
				a_down=A0[Index3D (nx, ny, i, j-1, k )];
			}
			else
			{
				a_down=sh_A0[iny-1][inx];
			}
			
			if(y_h_bound)
			{
				a_top=A0[Index3D (nx, ny, i, j+1, k )];
			}
			else
			{
				a_top=sh_A0[iny+1][inx];
			}
			
			//calculate results and write back
			Anext[Index3D (nx, ny, i, j, k)] = a_left + a_right + a_top + a_down +  top + bottom
			                                  - 6.0f * sh_A0[iny][inx]  / (fac*fac);;
		}
		__syncthreads();
		
		//swap data 
		//swap data from current to bottom
		//from top to current
		bottom=sh_A0[iny][inx];
		sh_A0[iny][inx]=top;
		__syncthreads();
	}
}



//For Optimization 3
__global__ void block2D_hybrid_coarsen_x(float fac,float *A0,float *Anext, int nx, int ny, int nz)
{

	//thread coarsening along x direction
	int inx=threadIdx.x;
	int iny=threadIdx.y;
	int inx2=threadIdx.x+blockDim.x;
	int i = blockIdx.x*blockDim.x*COARSEN_FACTOR+inx;
	int i2= blockIdx.x*blockDim.x*COARSEN_FACTOR+inx2;
    int j = blockIdx.y*blockDim.y+iny;
	
	//memory shared memory
	__shared__ float sh_A0[BLOCK_SIZE_Y/2][BLOCK_SIZE_X*2];
	sh_A0[iny][inx]=0.0f;
	sh_A0[iny][inx2]=0.0f;
	__syncthreads();

	//get reasonable region for load and store
	bool w_region =  i>0 && j>0 &&(i<nx-1) &&(j<ny-1) ;
	bool w_region2 =  j>0 &&(i2<nx-1) &&(j<ny-1) ;
	bool x_l_bound = (inx==0);
	bool x_h_bound = (inx2==(blockDim.x*2-1));
	bool y_l_bound = (iny==0);
	bool y_h_bound = (iny==(blockDim.y-1));
	
	
	//register for bottom and top planes
	//because of thread coarsening, we need to doulbe registers
	float bottom,bottom2,top,top2;

	//load data for bottom and current 
	bottom=A0[Index3D (nx, ny, i, j, 0)];
	bottom2=A0[Index3D (nx, ny, i2, j, 0)];
	sh_A0[iny][inx]=A0[Index3D (nx, ny, i, j, 1)];
	sh_A0[iny][inx2]=A0[Index3D (nx, ny, i2, j, 1)];
	__syncthreads();
	
	//inner loop for z direction
	for(int k=1;k<nz-1;k++)
	{

		//load data for top plane
		top=A0[Index3D (nx, ny, i, j, k+1)];
		top2=A0[Index3D (nx, ny, i2, j, k+1)];
		float a_left_right,a_top,a_down;
		float partial;
		
		if(w_region)
		{
			//get partial results
			partial=bottom - 6.0f * sh_A0[iny][inx]  / (fac*fac);
			
			//load required data on xy planes
			//if it on shared memory, load from shared memory
			//if not, load from global memory
			if(x_l_bound)
			{
				a_left_right=A0[Index3D (nx, ny, i-1, j, k )];
			}
			else
			{
				a_left_right=sh_A0[iny][inx-1];
			}

			if(y_l_bound)
			{
				a_down=A0[Index3D (nx, ny, i, j-1, k )];
			}
			else
			{
				a_down=sh_A0[iny-1][inx];
			}
			
			if(y_h_bound)
			{
				a_top=A0[Index3D (nx, ny, i, j+1, k )];
			}
			else
			{
				a_top=sh_A0[iny+1][inx];
			}
	
			Anext[Index3D (nx, ny, i, j, k)] = a_left_right + sh_A0[iny][inx+1] + a_top + a_down +  top + partial;
			

		}
		
		//do again for next virtual block
		if(w_region2)
		{
			
			partial=bottom2 - 6.0f * sh_A0[iny][inx2]  / (fac*fac);
			if(x_h_bound)
			{
				a_left_right=A0[Index3D (nx, ny, i2+1, j, k )];
			}
			else
			{
				a_left_right=sh_A0[iny][inx2+1];
			}

			if(y_l_bound)
			{
				a_down=A0[Index3D (nx, ny, i2, j-1, k )];
			}
			else
			{
				a_down=sh_A0[iny-1][inx2];
			}
			
			if(y_h_bound)
			{
				a_top=A0[Index3D (nx, ny, i2, j+1, k )];
			}
			else
			{
				a_top=sh_A0[iny+1][inx2];
			}
	
			Anext[Index3D (nx, ny, i2, j, k)] = a_left_right + sh_A0[iny][inx2-1] + a_top + a_down +  top2 +partial ;
			

		}
		
		__syncthreads();
		
		//swap data
		bottom=sh_A0[iny][inx];
		sh_A0[iny][inx]=top;
		bottom2=sh_A0[iny][inx2];
		sh_A0[iny][inx2]=top2;
		__syncthreads();
	}
}


__global__ void block2D_opt_1(float fac,float *A0,float *Anext, int nx, int ny, int nz)
{
	//For optimization 1
}



__global__ void block2D_opt_2(float fac,float *A0,float *Anext, int nx, int ny, int nz)
{
	//For optimization 2
	
}

__global__ void block2D_opt_3(float fac,float *A0,float *Anext, int nx, int ny, int nz)
{
	//For optimization 3
	
}
